#include "channelsort.h"
#include "sorter.h"

#include <vdr/tools.h>

#define CHANNELMARKOBSOLETE "OBSOLETE"
#define CHANNELMARKTEMPORARY "rovvisorio"
#define CHANNELMARKTEMPORARY2 "ROVVISORIO"
#define MAXVALIDLCN 925
#define INVALIDLCN 9999

cSorter *cSorter::s_Instance = NULL;

cSorter::cSorter() {
  s_Active = false;
}

cSorter::~cSorter() {
  if (s_Active)
    Stop();
  s_Instance = NULL;
  delete s_Instance;
}

void cSorter::Init(void) {
  if (s_Instance == NULL) {
    s_Instance = new cSorter;
    s_Instance->Start();
  }
}

void cSorter::Exit(void) {
    if (s_Instance != NULL) {
	s_Instance->Stop();
	DELETENULL(s_Instance);
    }
}

void cSorter::Restart(void) {
  Exit();
  Init();
}

void cSorter::Stop(void) {
    s_Active = false;
    Cancel(3);
}

void cSorter::Action(void) {
  isyslog("Channels sorting");
  s_Active = true;
  Sort();
}

bool cSorter::CheckInvalid(cChannel *ch)
{
  if (endswith(ch->Name(), CHANNELMARKOBSOLETE) || strcmp(ch->Name(), "") == 0 || strcmp(ch->Name(), "???") == 0 || strcmp(ch->Name(), ".") == 0)
    return true;
  return false;
}

int cSorter::Compare(cChannel *ch1, cChannel *ch2) {
  // if return
  // < 0 ch2 greater then ch1
  // = 0 ch1 equal ch2
  // > 0 ch1 greater then ch2
  if (ch1->GroupSep()) {
    if (strcmp(ch1->Name(), "Terr") == 0 && ch2->IsTerr())
      return -1;
    if (strcmp(ch1->Name(), "Sat") == 0 && ch2->IsSat())
      return -1;
    if (strcmp(ch1->Name(), "Cable") == 0 && ch2->IsCable())
      return -1;
    return 1;
  }
  int r = 0;
  for(int i = 0; SortMode[i] != 0 && i < 5; i++) {
    switch (SortMode[i]) {
      case 1: //type
        if (ch1->IsTerr() && ch2->IsSat())
          r = -1;
        if (ch1->IsSat() && ch2->IsTerr())
          r = 1;
        break;
      case 2: //frequency
        r = ch1->Frequency() - ch2->Frequency();
        break;
      case 3: //sid
        r = ch1->Sid() - ch2->Sid();
        break;
      case 4: //number
        r = ch1->Number() - ch2->Number();
        break;
      case 5: //name
        r = strcoll(ch1->Name(), ch2->Name());
        break;
      case 6: //provider
        r = strcoll(ch1->Provider(), ch2->Provider());
        break;
    }
    if (r != 0)
      return r;
  }
  return r;
}

void cSorter::Sort(void) {
  cStateKey WriteState;
  cChannels* Channels = (cChannels*) cChannels::GetChannelsWrite(WriteState, 30000);

  if (!Channels)
    return;

  bool sorted = false;
  
  cFixedChannels Fixed(AddDirectory("/etc/vdr/plugins/channelsort", "fixed.conf"));
  
  cVector<int> MaxChannelNumber(3); //0 = Terr; 1 = Sat; 2 = Cable
  
  cVector<bool> SepararorSet(3); //0 = Terr; 1 = Sat; 2 = Cable
  
  cVector<int> FoundLcn(Channels->Count());
  
  cVector<int> foundChannelsNumbers;
  
  if (AutoGroup) {
    for (cChannel *ch = Channels->First(); ch; ch = Channels->Next(ch)) {
      if (ch->GroupSep()) {
        cChannel *prev = Channels->Prev(ch);
        Channels->Del(ch);
        if (prev)
          ch = prev;
        else
          ch = Channels->First();
      }
    }
  }
  
  for (cChannel *ch = Channels->First(); ch; ch = Channels->Next(ch)) {
    if (!ch->GroupSep()) {
      if (DeleteInvalid && CheckInvalid(ch))
      {
        cChannel *prev = Channels->Prev(ch);
        Channels->Del(ch);
        sorted = true;
        if (prev)
          ch = prev;
        else
          ch = Channels->First();
        continue;
      }
      
      if (AutoGroup) {
        if (ch->IsTerr() && !SepararorSet[0]) {
          cChannel *sep = new cChannel();
          sep->SetGroupSep(true);
          sep->SetName("Terr", "", "");
          Channels->Add(sep);
          SepararorSet[0] = true;
        }
        if (ch->IsSat() && !SepararorSet[1]) {
          cChannel *sep = new cChannel();
          sep->SetGroupSep(true);
          sep->SetName("Sat", "", "");
          Channels->Add(sep);
          SepararorSet[1] = true;
        }
      }
      
      int fixedNumber = 0;
      
      //renumber from config file
      if (AdvancedConfig) {
        for (cChannel *fixedCh = Fixed.First(); fixedCh; fixedCh = Fixed.Next(fixedCh)) {
          if (strcmp(ch->Name(), fixedCh->Name()) == 0  &&( ch->Frequency() == fixedCh->Frequency() || ch->Frequency() == fixedCh->Frequency()+1 || ch->Frequency() == fixedCh->Frequency()-1 )) { 
            if (foundChannelsNumbers.AppendUnique(fixedCh->Rid())) {
              if (ch->IsTerr())
                fixedNumber = fixedCh->Rid() + FixedOffset[0];
              if (ch->IsSat())
                fixedNumber = fixedCh->Rid() + FixedOffset[1];
            }
            break;
          }
        }
      }

      if (fixedNumber != 0) {
        bool Clash = false;
        for (int i = 0; i < FoundLcn.Size(); ++i) {
          if (FoundLcn[i] == fixedNumber) {
            Clash = true;
            break;
          }
        }
        if (!Clash) {
          FoundLcn.Append(fixedNumber);
        } else {
          for (cChannel *clashCh = Channels->First(); clashCh; clashCh = Channels->Next(clashCh)) {
            if (clashCh->Number() == fixedNumber) {
              clashCh->SetNumber(INVALIDLCN);
              for (cChannel *compCh = Channels->Last(); compCh != clashCh; compCh = Channels->Prev(compCh)) {
                int comp = Compare(clashCh, compCh);
                if ((comp < 1 && Reverse == false) || (comp > 1 && Reverse == true)) {
                  cChannel *next = Channels->Next(clashCh);
                  if (compCh->Next())
                    Channels->Move(clashCh, compCh->Next());
                  else {
                    cChannel *temp = new cChannel();
                    Channels->Add(temp);
                    Channels->Move(clashCh, temp);
                    Channels->Del(temp);
                  }
                  sorted = true;
                  if (next)
                    clashCh = next;
                  else
                    clashCh = Channels->Last();
                  break;
                }
              }
              break;
            }
          }
        }
        ch->SetNumber(fixedNumber);
        if (ch->IsTerr() && fixedNumber > MaxChannelNumber[0])
          MaxChannelNumber[0] = fixedNumber;
        else if (ch->IsSat() && fixedNumber > MaxChannelNumber[1])
          MaxChannelNumber[1] = fixedNumber;
        else if (ch->IsCable() && fixedNumber > MaxChannelNumber[2])
          MaxChannelNumber[2] = fixedNumber;
      }
      else {
        //renumber LCN
        int Number = 0;
        cDvbTransponderParameters dtp(ch->Parameters());
        //Number = ch->Lcn();
        Number = dtp.Lcn();
        if (Number > 0 && Number <= MAXVALIDLCN) {
          bool Clash = false;
          for (int i = 0; i < FoundLcn.Size(); ++i) {
            if (FoundLcn[i] == Number) {
              Clash = true;
              break;
            }
          }
          if (!Clash) {
            FoundLcn.Append(Number);
            ch->SetNumber(Number);
            if (ch->IsTerr() && Number > MaxChannelNumber[0])
              MaxChannelNumber[0] = Number;
            else if (ch->IsSat() && Number > MaxChannelNumber[1])
              MaxChannelNumber[1] = Number;
            else if (ch->IsCable() && Number > MaxChannelNumber[2])
              MaxChannelNumber[2] = Number;
          } else {
            // Prioritize HD and not provisionary channels
            for (cChannel *clashCh = Channels->First(); clashCh; clashCh = Channels->Next(clashCh)) {
              if (clashCh->Number() == Number) {
                if ((strstr(ch->Name(), clashCh->Name()) && strstr(ch->Name(), "HD") && !strstr(clashCh->Name(), "HD")) || strstr(clashCh->Name(), CHANNELMARKTEMPORARY) || strstr(clashCh->Name(), CHANNELMARKTEMPORARY2)) {
                  clashCh->SetNumber(INVALIDLCN);
                  for (cChannel *compCh = Channels->Last(); compCh != clashCh; compCh = Channels->Prev(compCh)) {
                    int comp = Compare(clashCh, compCh);
                    if ((comp < 1 && Reverse == false) || (comp > 1 && Reverse == true)) {
                      cChannel *next = Channels->Next(clashCh);
                      if (compCh->Next())
                        Channels->Move(clashCh, compCh->Next());
                      else {
                        cChannel *temp = new cChannel();
                        Channels->Add(temp);
                        Channels->Move(clashCh, temp);
                        Channels->Del(temp);
                      }
                      sorted = true;
                      if (next)
                        clashCh = next;
                      else
                        clashCh = Channels->Last();
                      break;
                    }
                  ch->SetNumber(Number);
                  }
                } else {
                  ch->SetNumber(INVALIDLCN);
                }
                break;
              }
            }
          }
        } else
          ch->SetNumber(INVALIDLCN);
      }
    }

    for (cChannel *compCh = Channels->First(); compCh != ch; compCh = Channels->Next(compCh)) {
      int comp = Compare(ch, compCh);
      if ((comp < 0 && Reverse == false) || (comp > 0 && Reverse == true)) {
        cChannel *prev = Channels->Prev(ch);
        Channels->Move(ch, compCh);
        sorted = true;
        if (prev)
          ch = prev;
        else
          ch = Channels->First();
        break;
      }
    }
  }

  for (cChannel *ch = Channels->First(); ch; ch = Channels->Next(ch)) {
    if (ch->Number() == INVALIDLCN) {
      if (ch->IsTerr())
        ch->SetNumber(++MaxChannelNumber[0]);
      else if (ch->IsSat())
        ch->SetNumber(++MaxChannelNumber[1]);
      else if (ch->IsCable())
        ch->SetNumber(++MaxChannelNumber[2]);
      else 
        ch->SetNumber(++MaxChannelNumber[0]);
    }    
  }    

  if (sorted)
    Channels->SetModifiedByUser(); //Force channels saving

  WriteState.Remove();
  isyslog("Channels sorted");
}
